@@ -4,6 +4,7 @@ use strict;
 use warnings;
 
 use Future;
+use Scalar::Util qw( blessed );
 
 my $pending_request;
 my $pending_response;
@@ -12,6 +13,29 @@ my $pending_on_chunk;
 
 sub new { bless [], shift }
 
+sub _pull_content
+{
+   my ( $content, $request ) = @_;
+
+   if( !ref $content ) {
+      $request->add_content( $content );
+   }
+   elsif( ref $content eq "CODE" ) {
+      while( defined( my $chunk = $content->() ) ) {
+         _pull_content( $chunk, $request );
+      }
+   }
+   elsif( blessed $content and $content->isa( "Future" ) ) {
+      $content->on_done( sub {
+         my ( $chunk ) = @_;
+         _pull_content( $chunk, $request );
+      });
+   }
+   else {
+      die "TODO: Not sure how to handle $content";
+   }
+}
+
 sub do_request
 {
    shift; # self
@@ -23,14 +47,7 @@ sub do_request
    $pending_on_header = delete $args{on_header};
 
    if( my $request_body = delete $args{request_body} ) {
-      if( ref $request_body eq "CODE" ) {
-         while( defined( my $chunk = $request_body->() ) ) {
-            $pending_request->add_content( $chunk );
-         }
-      }
-      else {
-         $pending_request->content( $request_body );
-      }
+      _pull_content( $request_body, $pending_request );
    }
 
    delete $args{expect_continue};
